from cpk.exceptions import CPKException

from cpk.utils.progress_bar import ProgressBar
from ..cli import cpklogger

from ..types import Machine

DEFAULT_TCP_PORT = "2375"
DEFAULT_MACHINE = "unix:///var/run/docker.sock"
DEFAULT_REGISTRY = "docker.io"

DOCKER_INFO = """
Docker Endpoint:
  Machine: {machine}
  Hostname: {Name}
  Operating System: {OperatingSystem}
  Kernel Version: {KernelVersion}
  OSType: {OSType}
  Architecture: {Architecture}
  Total Memory: {MemTotal}
  CPUs: {NCPU}
"""


def push_image(machine: Machine, image: str, progress=True) -> str:
    client = machine.get_client()
    # keep track of total/pushed layers
    layers = set()
    pushed = set()
    pbar = ProgressBar() if progress else None
    final_digest = None
    for line in client.api.push(*image.split(":"), stream=True, decode=True):
        if "error" in line:
            msg = str(line["error"])
            msg = f"Cannot push image {image}:\n{msg}"
            raise CPKException(msg)

        if "aux" in line:
            if "Digest" in line["aux"]:
                final_digest = line["aux"]["Digest"]
                continue

        if "id" not in line:
            continue

        layer_id = line["id"]
        layers.add(layer_id)
        if line["status"] in ["Layer already exists", "Pushed"]:
            pushed.add(layer_id)
        # update progress bar
        if progress:
            percentage = max(0.0, min(1.0, len(pushed) / max(1.0, len(layers)))) * 100.0
            pbar.update(percentage)
    if progress:
        pbar.done()
    if final_digest is None:
        msg = "Expected to get final digest, but none arrived "
        cpklogger.warning(msg)
    return final_digest


def transfer_image(origin: Machine, destination: Machine, image, image_size):
    # TODO: re-implement this using in-Python sockets
    # monitor_info = "" if which("pv") else " (install `pv` to see the progress)"
    # cpklogger.info(f'Transferring image "{image}": [{origin}] -> [{destination}]'
    #                f'{monitor_info}...')
    # data_source = ["docker", "-H=%s" % origin, "save", image]
    # data_destination = ["docker", "-H=%s" % destination, "load"]
    # progress_monitor = ["|", "pv", "-cN", "image", "-s", image_size] if which("pv") else []
    # cmd = data_source + progress_monitor + data_destination
    # TODO: re-enable this
    # start_command_in_subprocess(cmd, nostdout=True)
    pass



import os
import re
import subprocess
from typing import Optional

from cpk.utils.misc import run_cmd
from cpk.types import GitRepositoryOrigin, GitRepositoryIndex, GitRepositoryVersion, GitRepository


def get_repo_info(path: str) -> GitRepository:
    path = os.path.join(path, '.git')
    if not os.path.exists(path):
        return GitRepository.default()
    # ---
    # get repo info
    # - sha
    try:
        sha = run_cmd(["git", "--git-dir", f'"{path}"',
                       "rev-parse", "HEAD"])[0]
    except subprocess.CalledProcessError as e:
        if e.returncode == 128:
            # empty repository
            return GitRepository.default()
    # - branch
    branch = run_cmd(["git", "--git-dir", f'"{path}"',
                      "rev-parse", "--abbrev-ref", "HEAD"])[0]
    # - remote
    origin_url = run_cmd(["git", "--git-dir", f'"{path}"',
                          "config", "--get", "remote.origin.url"])[0]
    if origin_url.endswith(".git"):
        origin_url = origin_url[:-4]
    if origin_url.endswith("/"):
        origin_url = origin_url[:-1]
    repo = origin_url.split("/")[-1]
    # - info about current git INDEX
    nmodified = len(
        run_cmd(["git", "--git-dir", f'"{path}"',
                 "status", "--porcelain", "--untracked-files=no"]))
    nadded = len(run_cmd(["git", "--git-dir", f'"{path}"', "status", "--porcelain"]))
    clean = abs(nmodified) + abs(nadded) == 0
    # - head/closest git tag
    head_tag = run_cmd(
        [
            "git",
            "--git-dir",
            f'"{path}"',
            "describe",
            "--exact-match",
            "--tags",
            "HEAD",
            "2>/dev/null",
            "||",
            ":",
        ]
    )
    head_tag = head_tag[0] if (head_tag and clean) else None
    closest_tag = run_cmd(["git", "--git-dir", f'"{path}"', "tag"])
    closest_tag = closest_tag[-1] if closest_tag else None
    # return info
    return GitRepository(
        name=repo,
        sha=sha,
        branch=branch,
        present=True,
        detached=branch == "HEAD",
        version=GitRepositoryVersion(
            head=head_tag,
            closest=closest_tag
        ),
        origin=GitRepositoryOrigin(
            url=origin_url,
            url_https=remote_url_to_https(origin_url),
            organization=remote_url_to_organization(origin_url)
        ),
        index=GitRepositoryIndex(
            clean=clean,
            num_added=nadded,
            num_modified=nmodified
        )
    )


def remote_url_to_https(remote_url: str) -> str:
    ssh_pattern = "git@([^:]+):([^/]+)/(.+)"
    res = re.search(ssh_pattern, remote_url, re.IGNORECASE)
    if res:
        return f"https://{res.group(1)}/{res.group(2)}/{res.group(3)}"
    return remote_url


def remote_url_to_organization(remote_url: str) -> Optional[str]:
    remote_url = remote_url_to_https(remote_url)
    https_pattern = "http[s]?://[^/]+/([^/]+)/.+"
    res = re.search(https_pattern, remote_url, re.IGNORECASE)
    if res:
        return res.group(1)
    return None



import argparse
import copy
import datetime
import json
import sys
import time
from typing import Optional

from docker.errors import ImageNotFound, APIError
from termcolor import colored

from cpk import CPKProject, cpkconfig
from cpk.utils.docker import transfer_image
from cpk.utils.misc import human_time, configure_binfmt
from .endpoint import CLIEndpointInfoCommand
from .info import CLIInfoCommand
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...constants import ARCH_TO_DOCKER_PLATFORM
from ...exceptions import CPKProjectBuildException
from ...types import Machine, Arguments
from ...utils.cli import check_git_status
from ...utils.image_analyzer import EXTRA_INFO_SEPARATOR, ImageAnalyzer, SEPARATORS_LENGTH
from ...utils.machine import get_machine


class CLIBuildCommand(AbstractCLICommand):

    KEY = 'build'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        parser.add_argument(
            "--pull",
            default=False,
            action="store_true",
            help="Whether to pull the latest base image used by the Dockerfile",
        )
        parser.add_argument(
            "--no-cache",
            default=False,
            action="store_true",
            help="Whether to use the Docker cache"
        )
        parser.add_argument(
            "--force-cache",
            default=False,
            action="store_true",
            help="Whether to force Docker to use an old version of the same image as cache",
        )
        parser.add_argument(
            "--no-multiarch",
            default=False,
            action="store_true",
            help="Whether to disable multiarch support (based on bin_fmt)",
        )
        parser.add_argument(
            "-A",
            "--build-arg",
            default=[],
            action="append",
            nargs=2,
            metavar=('key', 'value'),
            help="Build arguments to pass to Docker build",
        )
        parser.add_argument(
            "--push",
            default=False,
            action="store_true",
            help="Whether to push the resulting image"
        )
        parser.add_argument(
            "--rm",
            default=False,
            action="store_true",
            help="Remove the images once the build is finished (after pushing)",
        )
        parser.add_argument(
            "-b",
            "--base-tag",
            default=None,
            help="Docker tag for the base image. "
                 "Use when the base image is also a development version",
        )
        parser.add_argument(
            "--stamp",
            default=False,
            action="store_true",
            help="Stamp image with the build time"
        )
        # TODO: to be implemented
        parser.add_argument(
            "-D",
            "--destination",
            default=None,
            help="CPK machine or endpoint hostname where to deliver the image once built"
        )
        # TODO: to be implemented
        parser.add_argument(
            "--docs",
            default=False,
            action="store_true",
            help="Build the code documentation as well"
        )
        parser.add_argument(
            "--tag",
            default=None,
            type=str,
            help="Custom tag"
        )
        parser.add_argument(
            "--ncpus",
            default=None,
            type=int,
            help="Value to pass as build-arg `NCPUS` to docker build."
        )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        stime = time.time()

        # get project
        project = CPKProject(parsed.workdir, parsed=parsed)

        # show info about project
        CLIInfoCommand.execute(None, parsed)

        # check git workspace status
        proceed = check_git_status(project, parsed)
        if not proceed:
            return False

        # get info about docker endpoint
        CLIEndpointInfoCommand.execute(machine, parsed)

        # pick right value of `arch` given endpoint
        machine_arch = machine.get_architecture()
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine_arch
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # create docker client
        docker = machine.get_client()

        # define build-args
        buildargs = {"buildargs": {}, "labels": {}}
        # - add project build args
        buildargs["buildargs"].update({
            "ARCH": parsed.arch,
            "NAME": project.name,
            "DESCRIPTION": project.description,
            "ORGANIZATION": project.organization,
            "MAINTAINER": project.maintainer
        })
        # - add project labels
        buildargs["labels"].update(project.build_labels())
        # - build-arg NCPUS
        buildargs['buildargs']['NCPUS'] = \
            str(parsed.ncpus) if parsed.ncpus else str(machine.get_ncpus())

        # create defaults
        image = project.image(parsed.arch)

        # print info about multiarch
        cpklogger.info("Building an image for {} on {}.".format(parsed.arch, machine_arch))
        # - register bin_fmt in the target machine (if needed)
        if not parsed.no_multiarch:
            configure_binfmt(machine_arch, parsed.arch, docker, cpklogger)
        platform = ARCH_TO_DOCKER_PLATFORM.get(parsed.arch, None)

        # architecture target
        buildargs["buildargs"]["ARCH"] = parsed.arch

        # development base images
        if parsed.base_tag is not None:
            buildargs["buildargs"]["DISTRO"] = parsed.base_tag

        # custom build arguments
        for key, value in parsed.build_arg:
            buildargs["buildargs"][key] = value

        # cache
        if not parsed.no_cache:
            # check if the endpoint contains an image with the same name
            try:
                docker.images.get(image)
                is_present = True
            except (ImageNotFound, BaseException):
                is_present = False
            # ---
            if not is_present and parsed.pull:
                # try to pull the same image so Docker can use it as cache source
                cpklogger.info(f'Pulling image "{image}" to use as cache...')
                try:
                    machine.pull_image(image)
                    is_present = True
                except KeyboardInterrupt:
                    cpklogger.info("Aborting.")
                    return False
                except (ImageNotFound, BaseException):
                    cpklogger.warning(
                        f'An error occurred while pulling the image "{image}", maybe the '
                        "image does not exist"
                    )
            else:
                cpklogger.info("Found an image with the same name. Using it as cache source.")
            # configure cache
            if parsed.force_cache and is_present:
                buildargs["cache_from"] = [image]

        # stamp image
        build_time = "ND"
        if parsed.stamp:
            if project.is_dirty():
                cpklogger.error(
                    "Your git index is not clean. You can't stamp an image built "
                    "from a dirty index."
                )
                return False
            else:
                # project is clean
                build_time = None
                local_sha = project.version.sha
                # get remote image metadata
                try:
                    # TODO: parsed.machine is now NONE
                    labels = project.image_labels(parsed.machine, parsed.arch)
                    time_label = project.label("time")
                    sha_label = project.label("code.sha")
                    if time_label in labels and sha_label in labels:
                        remote_time = labels[time_label]
                        remote_sha = labels[sha_label]
                        if remote_sha == local_sha and remote_time != "ND":
                            cpklogger.debug("Identical image found. Reusing cache.")
                            # local and remote SHA match, reuse time
                            build_time = remote_time
                except BaseException as e:
                    cpklogger.warning(f"Cannot fetch image metadata. Reason: {str(e)}")
        # default build_time
        build_time = build_time or datetime.datetime.utcnow().isoformat()
        cpklogger.debug(f"Image timestamp: {build_time}")
        # add timestamp label
        buildargs["labels"][project.label("time")] = build_time

        # collect build args
        buildargs.update(
            {
                "path": project.path,
                "rm": True,
                "pull": parsed.pull,
                "nocache": parsed.no_cache,
                "tag": image,
                "platform": platform
            }
        )
        cpklogger.debug("Build arguments:\n%s\n" % json.dumps(buildargs, sort_keys=True, indent=4))

        # build image
        buildlog = []
        print("=" * SEPARATORS_LENGTH)
        try:
            for line in docker.api.build(**buildargs, decode=True):
                line = _build_line(line)
                if not line:
                    continue
                try:
                    sys.stdout.write(line)
                    buildlog.append(line)
                except UnicodeEncodeError:
                    pass
        except APIError as e:
            cpklogger.error(f"An error occurred while building the project image:\n{str(e)}")
            return False
        except CPKProjectBuildException:
            cpklogger.error(f"An error occurred while building the project image.")
            return False
        dimage = docker.images.get(image)

        # tag release images
        if project.is_release():
            rimage = project.image_release(parsed.arch, avoid_defaults=True)
            dimage.tag(*rimage.split(":"))
            msg = f"Successfully tagged {rimage}"
            buildlog.append(msg)
            print(msg)

        # TODO: fix this
        # # build code docs
        # if parsed.docs:
        #     docs_args = ["--quiet"] * int(not parsed.verbose)
        #     # build docs
        #     cpklogger.info("Building documentation...")

        # get image history
        historylog = [(layer["Id"], layer["Size"]) for layer in dimage.history()]

        # round up extra info
        extra_info = []
        # - launchers info
        launchers = project.launchers()
        if len(launchers) > 0:
            extra_info.append("Image launchers:")
            for launcher in sorted(launchers):
                extra_info.append(" - {:s}".format(launcher))
            extra_info.append(EXTRA_INFO_SEPARATOR)
        # - timing
        extra_info.append("Time: {}".format(human_time(time.time() - stime)))
        # - documentation
        extra_info.append(
            "Documentation: {}".format(
                colored("Built", "green") if parsed.docs else colored("Skipped", "yellow")
            )
        )
        # compile extra info
        extra_info = "\n".join(extra_info)
        # run docker image analysis
        print("=" * SEPARATORS_LENGTH + "\n")
        cpklogger.info("Analyzing the image...")
        print()
        _, _, final_image_size = ImageAnalyzer.process(
            buildlog, historylog, extra_info=extra_info
        )
        # pull image (if the destination is different from the builder machine)
        if parsed.destination and machine.base_url != parsed.destination:
            transfer_image(
                origin=machine,
                destination=get_machine(parsed.destination, cpkconfig.machines),
                image=image,
                image_size=final_image_size,
            )
        # perform push (if needed)
        if parsed.push:
            # call command `push`
            from .push import CLIPushCommand
            push_args = copy.deepcopy(parsed)
            push_args.release = False
            CLIPushCommand.execute(machine, push_args)

        # perform remove (if needed)
        if parsed.rm:
            from .clean import CLICleanCommand
            # noinspection PyBroadException
            try:
                # call command `clean`
                CLICleanCommand.execute(machine, copy.deepcopy(parsed))
            except BaseException:
                cpklogger.warn("We had some issues cleaning up the image. Just a heads up!")


def _build_line(line):
    if "error" in line and "errorDetail" in line:
        msg = line["errorDetail"]["message"]
        cpklogger.error(msg)
        raise CPKProjectBuildException(msg)
    if "stream" not in line:
        return None
    line = line["stream"].strip("\n")
    if not line:
        return None
    # this allows apps inside docker build to clear lines
    if not line.endswith("\r"):
        line += "\n"
    # ---
    return line



import argparse
from typing import Optional

from cpk import cpkconfig
from cpk.cli import AbstractCLICommand, cpklogger
from cpk.types import Machine, Arguments
from cpk.utils.misc import ask_confirmation


class CLIMachineRemoveCommand(AbstractCLICommand):

    KEY = 'machine remove'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent], add_help=False)
        parser.add_argument(
            'name',
            type=str,
            help="Name of the machine to remove"
        )
        # ---
        return parser

    @staticmethod
    def execute(_: Machine, parsed: argparse.Namespace) -> bool:
        # make sure the machine exists
        if parsed.name not in cpkconfig.machines:
            cpklogger.error(f"The machine '{parsed.name}' does not exist.")
            return False
        # get machine
        machine = cpkconfig.machines[parsed.name]
        # show some info
        cpklogger.info("The following machine will be deleted.")
        print(machine)
        # ask for confirmation
        granted = ask_confirmation(
            cpklogger,
            message=f"The machine '{parsed.name}' will be deleted. This cannot be undone.",
            default='n'
        )
        if not granted:
            raise KeyboardInterrupt()
        # remove machine
        machine.remove(cpklogger)
        cpklogger.info("Machine deleted.")



import argparse
from typing import Optional, Dict, Type

from cpk.cli import AbstractCLICommand
from cpk.cli.commands.machine.create import CLIMachineCreateCommand
from cpk.cli.commands.machine.info import CLIMachineInfoCommand
from cpk.cli.commands.machine.list import CLIMachineListCommand
from cpk.cli.commands.machine.remove import CLIMachineRemoveCommand
from cpk.types import Machine, Arguments

_supported_subcommands: Dict[str, Type[AbstractCLICommand]] = {
    "create": CLIMachineCreateCommand,
    "info": CLIMachineInfoCommand,
    "rm": CLIMachineRemoveCommand,
    "remove": CLIMachineRemoveCommand,
    "ls": CLIMachineListCommand,
    "list": CLIMachineListCommand,
}


class CLIMachineCommand(AbstractCLICommand):

    KEY = 'machine'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        # create a temporary parser used to select the subcommand
        parser = argparse.ArgumentParser(parents=[parent], prog='cpk machine')
        parser.add_argument(
            'subcommand',
            choices=_supported_subcommands.keys(),
            help=f"Subcommand. Can be any of {', '.join(_supported_subcommands.keys())}"
        )
        parsed, _ = parser.parse_known_args(args)
        # return subcommand's parser
        subcommand = _supported_subcommands[parsed.subcommand]
        return subcommand.parser(parser, args)

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        subcommand = _supported_subcommands[parsed.subcommand]
        return subcommand.execute(machine, parsed)



import argparse
from typing import Optional

from docker.errors import APIError

from .endpoint import CLIEndpointInfoCommand
from .info import CLIInfoCommand
from .. import AbstractCLICommand, cpklogger
from ... import CPKProject
from ...exceptions import CPKProjectPullException
from ...types import Machine, Arguments


class CLIPullCommand(AbstractCLICommand):

    KEY = 'pull'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        parser.add_argument(
            "--tag",
            default=None,
            type=str,
            help="Custom tag"
        )
        parser.add_argument(
            "--release",
            default=False,
            action="store_true",
            help="Pull release image as well",
        )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # get project
        project = CPKProject(parsed.workdir, parsed=parsed)

        # show info about project
        CLIInfoCommand.execute(machine, parsed)

        # get info about docker endpoint
        CLIEndpointInfoCommand.execute(machine, parsed)

        # pick right value of `arch` given endpoint
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine.get_architecture()
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # create defaults
        images = [project.image(parsed.arch)]

        # release image
        if parsed.release and project.is_release():
            images += [project.image_release(parsed.arch)]

        for image in images:
            # print info about registry
            msg = "Pulling image {} from {}.".format(image, project.registry)
            cpklogger.info(msg)
            # pull image
            try:
                machine.pull_image(image, progress=True)
            except APIError as e:
                cpklogger.error(f"An error occurred while pulling the project image:\n{str(e)}")
                return False
            except CPKProjectPullException:
                cpklogger.error(f"An error occurred while pulling the project image.")
                return False

        cpklogger.info("Image pulled successfully!")



import argparse
from typing import Optional, Dict, Type

from cpk.cli import AbstractCLICommand
from cpk.cli.commands.endpoint.info import CLIEndpointInfoCommand
from cpk.types import Machine, Arguments

_supported_subcommands: Dict[str, Type[AbstractCLICommand]] = {
    "info": CLIEndpointInfoCommand,
}


class CLIEndpointCommand(AbstractCLICommand):

    KEY = 'endpoint'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        # create a temporary parser used to select the subcommand
        parser = argparse.ArgumentParser(parents=[parent], prog='cpk endpoint')
        parser.add_argument(
            'subcommand',
            choices=_supported_subcommands.keys(),
            help=f"Subcommand. Can be any of {', '.join(_supported_subcommands.keys())}"
        )
        parsed, _ = parser.parse_known_args(args)
        # return subcommand's parser
        subcommand = _supported_subcommands[parsed.subcommand]
        return subcommand.parser(parser, args)

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        subcommand = _supported_subcommands[parsed.subcommand]
        return subcommand.execute(machine, parsed)



from typing import Optional

from .generic import ProjectVersion
from .null import NullAdapter
from ..utils.git import get_repo_info


class GitRepositoryAdapter(NullAdapter):

    def __init__(self, path: str):
        super(GitRepositoryAdapter, self).__init__('git')
        self._path = path
        self._repo = get_repo_info(self._path)
        self.enabled = self._repo.present

    @property
    def name(self) -> str:
        return self._repo.name

    @property
    def organization(self) -> str:
        return self._repo.origin.organization

    @property
    def version(self) -> ProjectVersion:
        if not self.enabled:
            return super(GitRepositoryAdapter, self).version
        # ---
        return ProjectVersion(
            tag=self._repo.branch,
            head=self._repo.version.head,
            closest=self._repo.version.closest,
            sha=self._repo.sha
        )

    @property
    def maintainer(self) -> Optional[str]:
        # TODO: compile maintainer in the form "First Last (Email)" from git
        return None



import argparse
import os
import re
import subprocess
from shutil import which
from typing import Optional, Callable

import cpk
from cpk.utils.misc import configure_binfmt
from .endpoint import CLIEndpointInfoCommand
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...types import DockerImageName, Machine, Arguments


class CLIDecorateCommand(AbstractCLICommand):
    KEY = 'decorate'

    DOCKER_IMAGE_REGEX = r"^(([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])\.)*" \
                         r"([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])" \
                         r"(:[0-9]+\/)?([0-9a-z-]+[/@]?)([0-9a-z-]+)" \
                         r"[/@]?([0-9a-z-]+)?(:[a-z0-9\.-]+)?$"
    EMAIL_ADDRESS_REGEX = r"^[a-z0-9]+[\._]?[a-z0-9]+[@]\w+[.]\w{2,3}$"
    DOCKER_IMAGE_ERROR_MSG = "The pattern must be: [registry/]owner/repository[:tag]"
    EMAIL_ADDRESS_ERROR_MSG = "The pattern must be: 'user@domain.tld'"

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        # ---
        parser.add_argument(
            nargs=1,
            dest="input",
            metavar="INPUT",
            help="The input image",
            type=regex_type(CLIDecorateCommand.DOCKER_IMAGE_REGEX,
                            CLIDecorateCommand.DOCKER_IMAGE_ERROR_MSG)
        )
        parser.add_argument(
            nargs=1,
            dest="output",
            metavar="OUTPUT",
            help="The output image",
            type=regex_type(CLIDecorateCommand.DOCKER_IMAGE_REGEX,
                            CLIDecorateCommand.DOCKER_IMAGE_ERROR_MSG)
        )
        parser.add_argument(
            "-m",
            "--maintainer",
            required=True,
            help="The image maintainer's name",
            type=str
        )
        parser.add_argument(
            "-e",
            "--email",
            default=None,
            help="The image maintainer's email address",
            type=regex_type(CLIDecorateCommand.EMAIL_ADDRESS_REGEX,
                            CLIDecorateCommand.EMAIL_ADDRESS_ERROR_MSG)
        )
        parser.add_argument(
            "--no-multiarch",
            default=False,
            action="store_true",
            help="Whether to disable multiarch support (based on bin_fmt)",
        )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # pick right value of `arch` given endpoint
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine.get_architecture()
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # parse `input`
        input_image = DockerImageName.from_image_name(parsed.input[0])
        cpklogger.debug(f"+ Input Image:\n{str(input_image)}")

        # parse `output`
        output_image = DockerImageName.from_image_name(parsed.output[0])

        # append arch to the end of the output image
        if output_image.arch is None:
            output_image.arch = parsed.arch
        cpklogger.debug(f"+ Output Image:\n{str(output_image)}")

        # make sure the `docker` CLI tool is installed
        docker_cli = which('docker')
        if docker_cli is None:
            cpklogger.error("The Docker CLI must be installed for this command to work.")
            return False
        cpklogger.debug(f"Docker CLI found at {docker_cli}")

        # find decorator project inside the `cpk` library
        cpk_dir = os.path.dirname(os.path.abspath(cpk.__file__))
        decorator_dir = os.path.join(cpk_dir, "decorator")
        dockerfile = os.path.join(decorator_dir, "Dockerfile")

        # compile maintainer string
        maintainer = f"{parsed.maintainer} ({parsed.email})" if parsed.email else parsed.maintainer

        # get info about docker endpoint
        CLIEndpointInfoCommand.execute(machine, parsed)

        # pick right value of `arch` given endpoint
        machine_arch = machine.get_architecture()
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine_arch
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # create docker client
        docker = machine.get_client()

        # print info about multiarch
        cpklogger.info("Decorating an image for {} on {}.".format(parsed.arch, machine_arch))
        # - register bin_fmt in the target machine (if needed)
        if not parsed.no_multiarch:
            configure_binfmt(machine_arch, parsed.arch, docker, cpklogger)

        # compile command
        cmd = [
            "docker", "build",
            # TODO: machine is not used here
            "-t", output_image.compile(),
            "-f", dockerfile,
            "--build-arg", f"ARCH={parsed.arch}",
            "--build-arg", f"BASE_REGISTRY={input_image.registry.compile(allow_defaults=True)}",
            "--build-arg", f"BASE_ORGANIZATION={input_image.user}",
            "--build-arg", f"BASE_REPOSITORY={input_image.repository}",
            "--build-arg", f"BASE_TAG={input_image.tag}",
            "--build-arg", f"ORGANIZATION={output_image.user}",
            "--build-arg", f"NAME={output_image.repository}",
            "--build-arg", f"MAINTAINER={maintainer}",
            "--build-arg", f"CPK_VERSION={cpk.__version__}",
            decorator_dir
        ]
        cpklogger.info(f"Decorating [{input_image.compile()}] -> [{output_image.compile()}]...")
        cpklogger.debug(f"Running command:\n\t{cmd}")

        # build image
        try:
            subprocess.check_call(cmd)
        except subprocess.SubprocessError as e:
            cpklogger.error(str(e))
            return False

        # success
        cpklogger.info(f"The given image was successfully decorated for CPK.\n"
                       f"Your CPK-compatible image is called\n\n"
                       f"\t\t{output_image.compile()}\n\n"
                       f"You can now use it as a base for your CPK projects and templates.")
        # ---
        return True


def regex_type(pattern: str, error_msg: str) -> Callable[[str], str]:
    def _validator(value: str) -> str:
        if not re.compile(pattern).match(value):
            raise argparse.ArgumentTypeError(f"\n\t{error_msg}")
        return value

    # ---
    return _validator



import ipaddress
import logging
import os
import subprocess
from pathlib import Path
from typing import Union, Optional

import docker
import yaml
from docker.errors import ContainerError, ImageNotFound, APIError

from cpk.constants import CANONICAL_ARCH, CONTAINER_LABEL_DOMAIN, BUILD_COMPATIBILITY_MAP


def run_cmd(cmd):
    cmd = " ".join(cmd)
    lines = subprocess.check_output(cmd, shell=True).decode("utf-8").split("\n")
    return list(filter(lambda line: len(line) > 0, lines))


def assert_canonical_arch(arch):
    if arch not in CANONICAL_ARCH.values():
        raise ValueError(
            f"Given architecture {arch} is not supported. "
            f"Valid choices are: {', '.join(list(set(CANONICAL_ARCH.values())))}"
        )


def canonical_arch(arch):
    if arch not in CANONICAL_ARCH:
        raise ValueError(
            f"Given architecture {arch} is not supported. "
            f"Valid choices are: {', '.join(list(set(CANONICAL_ARCH.values())))}"
        )
    # ---
    return CANONICAL_ARCH[arch]


def cpk_label(key, value=None):
    label = f"{CONTAINER_LABEL_DOMAIN}.{key.lstrip('.')}"
    if value is not None:
        label = f"{label}={value}"
    return label


def sanitize_hostname(hostname: str) -> str:
    try:
        ipaddress.ip_address(hostname)
        return hostname
    except ValueError:
        return f"{hostname}.local" if "." not in hostname else hostname


def parse_configurations(config_file: str) -> dict:
    with open(config_file, "rt") as fin:
        # TODO: use safe_load instead
        configurations_content = yaml.load(fin, Loader=yaml.SafeLoader)
    if "version" not in configurations_content:
        raise ValueError("The configurations file must have a root key 'version'.")
    # TODO: handle configuration schemas properly (i.e., using JSON/XML schemas)
    if configurations_content["version"] == "1.0":
        return configurations_content["configurations"]


def configure_binfmt(machine_arch: str, arch: str, epoint: docker.DockerClient, logger):
    machine_arch = CANONICAL_ARCH[machine_arch]
    compatible_archs = BUILD_COMPATIBILITY_MAP[machine_arch]
    if arch not in compatible_archs:
        logger.info("Configuring machine for multiarch...")
        try:
            epoint.containers.run(
                "multiarch/qemu-user-static:register",
                remove=True,
                privileged=True,
                command="--reset",
            )
            logger.info("Multiarch Enabled!")
        except (ContainerError, ImageNotFound, APIError) as e:
            msg = "Multiarch cannot be enabled on the target machine. " \
                  "This might create issues."
            logger.warning(msg)
            logger.debug(f"The error reads:\n\t{str(e)}\n")
    else:
        msg = "Working with an `{}` image on `{}`. Multiarch not needed!".format(
            arch, machine_arch
        )
        logger.info(msg)


def human_size(value: Union[int, float], suffix: str = "B", precision: int = 2):
    fmt = f"%3.{precision}f %s%s"
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(value) < 1024.0:
            return fmt % (value, unit, suffix)
        value /= 1024.0
    return fmt.format(value, "Yi", suffix)


def human_time(time_secs: Union[int, float], compact: bool = False):
    label = lambda s: s[0] if compact else " " + s
    days = int(time_secs // 86400)
    hours = int(time_secs // 3600 % 24)
    minutes = int(time_secs // 60 % 60)
    seconds = int(time_secs % 60)
    parts = []
    if days > 0:
        parts.append("{}{}".format(days, label("days")))
    if days > 0 or hours > 0:
        parts.append("{}{}".format(hours, label("hours")))
    if days > 0 or hours > 0 or minutes > 0:
        parts.append("{}{}".format(minutes, label("minutes")))
    parts.append("{}{}".format(seconds, label("seconds")))
    return ", ".join(parts)


def configure_ssh_for_cpk(logger: Optional[logging.Logger] = None):
    from cpk import cpkconfig
    user_home = Path.home()
    # create .cpk/ssh pool directory if it does not exist
    cpk_ssh_pool_dir = os.path.join(cpkconfig.path, "ssh")
    if logger:
        logger.debug(f"Creating directory '{cpk_ssh_pool_dir}'...")
        if os.path.exists(cpk_ssh_pool_dir):
            logger.debug(f"Directory '{cpk_ssh_pool_dir}' already exists, skipping.")
    os.makedirs(cpk_ssh_pool_dir, mode=0o700, exist_ok=True)
    # create .ssh dir if it does not exist
    ssh_fpath = os.path.join(user_home, ".ssh")
    if logger:
        logger.debug(f"Creating directory '{ssh_fpath}'...")
        if os.path.exists(ssh_fpath):
            logger.debug(f"Directory '{ssh_fpath}' already exists, skipping.")
    os.makedirs(ssh_fpath, mode=0o700, exist_ok=True)
    # read config file
    ssh_config_fpath = os.path.join(ssh_fpath, "config")
    ssh_config = []
    if os.path.isfile(ssh_config_fpath):
        if logger:
            logger.debug(f"Reading SSH configuration file '{ssh_config_fpath}'...")
        with open(ssh_config_fpath, "rt") as fin:
            ssh_config = fin.readlines()
    else:
        if logger:
            logger.debug(f"SSH configuration file '{ssh_config_fpath}' not found. Creating one...")
    # check if it is already configured
    found = False
    include_line = f"Include {os.path.join(cpk_ssh_pool_dir, '*')}"
    for line in ssh_config:
        if line.strip() == include_line:
            found = True
    if found:
        if logger:
            logger.debug(f"User SSH environment already configured for CPK.")
        return
    # add Include statement to the beginning of the file
    ssh_config = [
        "# SSH configuration for cpk (https://github.com/afdaniele/cpk)\n",
        f"{include_line}\n\n\n"
    ] + ssh_config
    # write ssh config file back to disk
    with open(ssh_config_fpath, "wt") as fout:
        if logger:
            logger.debug(f"Writing to SSH configuration file '{ssh_config_fpath}'...")
        fout.writelines(ssh_config)
        if logger:
            logger.debug(f"SSH configuration file '{ssh_config_fpath}' configured for CPK.")


def ask_confirmation(logger, message, default="y", question="Do you confirm?", choices=None):
    binary_question = False
    if choices is None:
        choices = {"y": "Yes", "n": "No"}
        binary_question = True
    choices_str = " ({})".format(", ".join([f"{k}={v}" for k, v in choices.items()]))
    default_str = f" [{default}]" if default else ""
    while True:
        logger.warn(f"{message.rstrip('.')}.")
        r = input(f"{question}{choices_str}{default_str}: ")
        if r.strip() == "":
            r = default
        r = r.strip().lower()
        if binary_question:
            if r in ["y", "yes", "yup", "yep", "si", "aye"]:
                return True
            elif r in ["n", "no", "nope", "nay"]:
                return False
        else:
            if r in choices:
                return r



import argparse
import os
import shutil
import subprocess
from pathlib import Path
from typing import Optional, Dict

from cpk.machine import SSHMachine
from docker.errors import ImageNotFound

from xdocker import get_configuration

from cpk import CPKProject
from cpk.utils.misc import configure_binfmt
from .endpoint import CLIEndpointInfoCommand
from .info import CLIInfoCommand
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...exceptions import NotACPKProjectException
from ...types import CPKFileMappingTrigger, Machine, Arguments, CPKFileMapping
from ...utils.cli import check_git_status

SUPPORTED_SUBCOMMANDS = [
    "attach"
]
RSYNC_DESTINATION_PATH = "/tmp/"


class CLIRunCommand(AbstractCLICommand):

    KEY = 'run'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        parser.add_argument(
            "-n",
            "--name",
            default=None,
            help="Name of the container"
        )
        parser.add_argument(
            "-c",
            "--cmd",
            default=None,
            help="Command to run in the Docker container"
        )
        parser.add_argument(
            "--pull",
            default=False,
            action="store_true",
            help="Whether to pull the latest image available",
        )
        parser.add_argument(
            "--force-pull",
            default=False,
            action="store_true",
            help="Whether to force pull the image of the project",
        )
        # parser.add_argument(
        #     "--plain",
        #     default=False,
        #     action="store_true",
        #     help="Whether to run the image without default module configuration",
        # )
        parser.add_argument(
            "--no-multiarch",
            default=False,
            action="store_true",
            help="Whether to disable multiarch support (based on bin_fmt)",
        )
        parser.add_argument(
            "-M",
            "--mount",
            default=False,
            const=True,
            action="store",
            nargs="?",
            type=str,
            help="Whether to mount the current project into the container. "
                 "Pass a comma-separated list of paths to mount multiple projects",
        )
        parser.add_argument(
            "--keep",
            default=False,
            action="store_true",
            help="Whether to NOT remove the container once stopped"
        )
        parser.add_argument(
            "-L",
            "--launcher",
            default=None,
            help="Launcher to invoke inside the container",
        )
        parser.add_argument(
            "-A",
            "--argument",
            dest="arguments",
            default=[],
            action="append",
            help="Arguments for the container command",
        )
        parser.add_argument(
            "--runtime",
            default="docker",
            type=str,
            help="Docker runtime to use to run the container"
        )
        parser.add_argument(
            "-X",
            dest="use_x_docker",
            default=False,
            action="store_true",
            help="Add X-Forwarding configuration (adds support for GUI applications)",
        )
        parser.add_argument(
            "-s",
            "--sync",
            default=False,
            action="store_true",
            help="Sync code from local project to remote"
        )
        parser.add_argument(
            "-sm",
            "--sync-mirror",
            default=False,
            action="store_true",
            help="Mirror code from local project to remote"
        )
        parser.add_argument(
            "--net", "--network_mode",
            dest="network_mode",
            default=None,
            type=str,
            help="Docker network mode"
        )
        parser.add_argument(
            "-d",
            "--detach",
            default=False,
            action="store_true",
            help="Detach from the container and let it run"
        )
        parser.add_argument(
            "--tag",
            default=None,
            type=str,
            help="Custom tag"
        )
        parser.add_argument(
            "docker_args",
            nargs="*",
            default=[]
        )
        # parser.add_argument(
        #     "subcommand",
        #     nargs="?",
        #     default=None,
        #     choices=SUPPORTED_SUBCOMMANDS,
        #     help="(Optional) Subcommand to execute"
        # )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # get project
        project = CPKProject(parsed.workdir, parsed=parsed)

        # show info about project
        CLIInfoCommand.execute(machine, parsed)

        # check git workspace status
        mount_source = parsed.mount is True or isinstance(parsed.mount, str)
        proceed = check_git_status(project, parsed, must_be_clean=mount_source)
        if not proceed:
            return False

        # get info about docker endpoint
        CLIEndpointInfoCommand.execute(machine, parsed)

        # pick right value of `arch` given endpoint
        machine_arch = machine.get_architecture()
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine_arch
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # create docker client
        docker = machine.get_client()

        # create defaults
        image = project.image(parsed.arch)
        parsed.name = parsed.name or f"cpk-run-{project.name.replace('/', '-')}"

        # subcommand "attach"
        # TODO: this will not work with Machine created from env, the host will be None
        # if parsed.subcommand == "attach":
        #     cpklogger.info(f"Attempting to attach to container '{parsed.name}'...")
        #     # run
        #     _run_cmd(
        #         [
        #             "docker",
        #             "-H=%s" % machine.host,
        #             "exec",
        #             "-it",
        #             parsed.name,
        #             "/entrypoint.sh",
        #             "bash",
        #         ],
        #         suppress_errors=True,
        #     )
        #     return True

        # environment
        environment = []
        # volumes
        volumes = []
        # module configuration
        module_configuration_args = []
        # active triggers
        triggers = {CPKFileMappingTrigger.DEFAULT.value}

        # trigger run:mount
        if mount_source:
            triggers.update({CPKFileMappingTrigger.RUN_MOUNT.value})

        # trigger launcher:<X>
        if parsed.launcher:
            triggers.update({f"launcher:{parsed.launcher}"})
        cpklogger.debug(f"Active triggers: {triggers}")

        # print info about multiarch
        cpklogger.info("Running an image for {} on {}.".format(parsed.arch, machine_arch))
        # - register bin_fmt in the target machine (if needed)
        if not parsed.no_multiarch:
            configure_binfmt(machine_arch, parsed.arch, docker, cpklogger)

        # x-docker configuration
        if parsed.use_x_docker:
            xconfig = get_configuration(docker)
            # TODO: once we switch to the DOcker SDK for Python, we should use this directly
            if "environment" in xconfig:
                for ekey, evalue in xconfig["environment"].items():
                    environment += ["-e", f"{ekey}={evalue}"]
            if "volumes" in xconfig:
                for vsrc, vdsc in xconfig["volumes"].items():
                    volumes += ["-v", "{:s}:{:s}:{:s}".format(vsrc, vdsc["bind"], vdsc["mode"])]
            if "runtime" in xconfig:
                module_configuration_args.append(f"--runtime={xconfig['runtime']}")

        # check runtime
        if shutil.which(parsed.runtime) is None:
            raise ValueError('Docker runtime binary "{}" not found!'.format(parsed.runtime))

        # network mode
        if parsed.network_mode is not None:
            module_configuration_args.append(f"--net={parsed.network_mode}")

        # mount source code (if requested)
        if mount_source:
            projects_to_mount = []
            # (always) mount current project
            paths_to_mount = [parsed.workdir] if parsed.mount is True else []
            # mount secondary projects
            if isinstance(parsed.mount, str):
                paths_to_mount.extend(
                    [os.path.join(os.getcwd(), p.strip()) for p in parsed.mount.split(",")]
                )
            # create mount points definitions
            for project_path in paths_to_mount:
                # make sure that the project exists
                if not os.path.isdir(project_path):
                    cpklogger.error('The path "{:s}" is not a CPK project'.format(project_path))
                    return False
                # get project info
                try:
                    proj = CPKProject(project_path)
                except NotACPKProjectException:
                    cpklogger.error(f"The path '{project_path}' does not contain a CPK project.")
                    return False
                projects_to_mount.append(proj)
        else:
            projects_to_mount = [project]

        # sync
        sync_remote = parsed.sync or parsed.sync_mirror
        if sync_remote:
            # does not make sense to rsync without mounting
            if not mount_source:
                cpklogger.error("The options -s/--sync, -sm/--sync-mirror can only be used "
                                "together with -M/--mount")
                return False
            # only allowed when mounting remotely
            if machine.is_local:
                cpklogger.error("The options -s/--sync, -sm/--sync-mirror can only be used "
                                "together with -H/--machine")
                return False
            # only allowed with SSH machines
            if not isinstance(machine, SSHMachine):
                cpklogger.error("The options -s/--sync, -sm/--sync-mirror can only be used "
                                "with SSH-based machines")
                return False
            # make sure rsync is installed
            if shutil.which("rsync") is None:
                cpklogger.error("The options -s/--sync, -sm/--sync-mirror requires the 'rsync' "
                                "tool. Please, install it and retry.")
                return False

        # mountpoints
        mountpoints: Dict[str, str] = {}
        rsync_mountpoints: Dict[str, str] = {}

        def _mount(_proj: CPKProject, _mapping: CPKFileMapping):
            # mountapoint source on the local machine (never depends on rsync)
            local_source_path = _proj.path
            local_mountpoint_source = _mapping.source if os.path.isabs(_mapping.source) else \
                os.path.join(local_source_path, _mapping.source)
            # mountpoint source to pass to docker (depends on rsync)
            source_path = _proj.path if not sync_remote else \
                os.path.join(RSYNC_DESTINATION_PATH, os.path.abspath(_proj.path).lstrip("/"))
            mountpoint_source = _mapping.source if os.path.isabs(_mapping.source) else \
                os.path.join(source_path, _mapping.source)
            # mountpoint destination (never depends on rsync)
            mountpoint_destination = _mapping.destination
            # mountpoint mode (i.e., ro, rw, etc)
            mode = _mapping.mode
            # resolve paths
            local_mountpoint_source = str(Path(local_mountpoint_source).resolve())
            mountpoint_source = str(Path(mountpoint_source).resolve())
            mountpoint_destination = str(Path(mountpoint_destination).resolve())
            # make sure we have no conflicts of mountpoints
            conflict = mountpoints.get(mountpoint_destination, None)
            if conflict not in [None, local_mountpoint_source]:
                cpklogger.error(f"Mountpoint '{mountpoint_destination}' inside the container "
                                f"wants to be claimed by both '{conflict}' and "
                                f"'{local_mountpoint_source}'.")
                return False
            # compile mounpoints and add them to list
            volumes.extend([
                "-v", "{:s}:{:s}:{:s}".format(mountpoint_source, mountpoint_destination, mode)
            ])
            mountpoints[mountpoint_destination] = local_mountpoint_source
            # only non-absolute paths can be rsynced
            if not os.path.isabs(_mapping.source):
                rsync_mountpoints[mountpoint_source] = local_mountpoint_source
            # ---
            return True

        # mappings
        for proj in projects_to_mount:
            # iterate over list of mappings
            for mapping in proj.mappings:
                if len(triggers.intersection(mapping.triggers)) <= 0:
                    continue
                # ---
                success = _mount(proj, mapping)
                if not success:
                    return False

        # sync
        if sync_remote:
            cpklogger.info(f"Syncing mountpoints...")
            remote_base = f"{machine.user}@{machine.host}"
            for rsync_destination, rsync_source in rsync_mountpoints.items():
                # rsync options
                rsync_options = "" if not parsed.sync_mirror else "--delete"
                rsync_options += f" --rsync-path=\"mkdir -p {rsync_destination} && rsync\""
                # run rsync
                remote_path = f"{remote_base}:{rsync_destination}".rstrip('/') + '/'
                cpklogger.info(f"Syncing mountpoint [{rsync_source}] -> [{remote_path}]")
                cmd = f"rsync --archive {rsync_options} {rsync_source}/ {remote_path}"
                _run_cmd(cmd, shell=True)
            cpklogger.info(f"Mountpoints synced!")

        # pulling image (if requested)
        if parsed.pull or parsed.force_pull:
            present = False
            try:
                docker.images.get(image)
                present = True
            except ImageNotFound:
                pass
            if present and not parsed.force_pull:
                cpklogger.info("Found an image with the same name. Use --force-pull to pull again")
            else:
                cpklogger.info('Pulling image "%s"...' % image)
                machine.pull_image(image)

        # cmd option
        if parsed.cmd and parsed.launcher:
            raise ValueError("You cannot use the option --launcher together with --cmd.")

        # launcher
        if parsed.launcher:
            environment.extend(["-e", f"CPK_LAUNCHER={parsed.launcher}"])

        cmd_option = [] if not parsed.cmd else [parsed.cmd]
        cmd_arguments = [] if not parsed.arguments else ["--"]

        for argval in parsed.arguments:
            cmd_arguments += f"--{argval}".split("=")

        # endpoint arguments
        docker_epoint_args = []
        # TODO: this should be using the Python SDK for Docker, not Docker CLI
        if machine.base_url is not None:
            docker_epoint_args += ["-H", machine.base_url]

        # docker arguments
        if not parsed.docker_args:
            parsed.docker_args = []
        if not parsed.keep:
            parsed.docker_args += ["--rm"]
        if parsed.detach:
            parsed.docker_args += ["-d"]

        # add container name to docker args
        parsed.docker_args += ["--name", parsed.name]
        # escape spaces in arguments
        parsed.docker_args = [a.replace(" ", "\\ ") for a in parsed.docker_args]

        # run
        exitcode = _run_cmd(
            [parsed.runtime]
            + docker_epoint_args
            + ["run", "-it"]
            + module_configuration_args
            + environment
            + parsed.docker_args
            + volumes
            + [image]
            + cmd_option
            + cmd_arguments,
            suppress_errors=True,
            return_exitcode=True
        )
        if parsed.detach:
            cpklogger.info("Your container is running in detached mode!")
        else:
            cpklogger.debug(f"Command exited with exit code [{exitcode}].")


def _run_cmd(cmd, get_output=False, print_output=False, suppress_errors=False, shell=False,
             return_exitcode=False):
    if shell and isinstance(cmd, (list, tuple)):
        cmd = " ".join([str(s) for s in cmd])
    cpklogger.debug("$ %s" % cmd)
    if get_output:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=shell)
        proc.wait()
        if proc.returncode != 0:
            if not suppress_errors:
                msg = "The command {} returned exit code {}".format(cmd, proc.returncode)
                cpklogger.error(msg)
                raise RuntimeError(msg)
        out = proc.stdout.read().decode("utf-8").rstrip()
        if print_output:
            print(out)
        return out
    else:
        if return_exitcode:
            res = subprocess.run(cmd, shell=shell)
            return res.returncode
        else:
            try:
                subprocess.check_call(cmd, shell=shell)
            except subprocess.CalledProcessError as e:
                if not suppress_errors:
                    raise e
